#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#ifndef WIN32
#include <signal.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <unistd.h>
#endif


#include <libTM.h>

#define MAXBUFSIZE (1024 * 1024)

void            do_error(uint32_t, int32_t);
void            usage(const char *);
#ifndef WIN32
void            sig_winch(int);
#endif

static int gWidth = 80;
static int show_cksum = 0;

/* TODO
 * Bundle all the writes into a buffer and write the complete buffer at the
 * end of the loop. There are too many tiny writes.
 */
int
main(int argc, char *argv[])
{
    char           *mesg, *stream, *progname;
    char           *s;
    int             j, bytes_of_meat, show_all_meat = 0;
    int32_t         chksum;
    int32_t        *intbuf;
    int32_t         err, numvals;
    uint32_t        sid, count = 0;
    double         *doublebuf;
#ifndef WIN32
    struct winsize  size;
#endif

    mesg = malloc(MAXBUFSIZE);
    stream = malloc(MAXBUFSIZE);
    progname = malloc(MAXBUFSIZE);
    intbuf = malloc(MAXBUFSIZE * sizeof(int32_t));
    doublebuf = malloc(MAXBUFSIZE * sizeof(double));

    strcpy(stream, "realtime");
    strcpy(progname, *argv);

    while (--argc > 0 && (*++argv)[0] == '-') {
        for (s = argv[0] + 1; *s != '\0'; s++) {
            switch (*s) {
            case 'h':
                usage(progname);
                exit(0);
                break;
            case 'c':
                strcpy(stream, "chooser");
                break;
            case 'k':
                show_cksum = 1;
                break;
            case 'a':
                show_all_meat = 1;
                break;
            default:
                usage(progname);
                exit(0);
                break;
            }
        }
    }

    if (argc < 3)
    {
        usage(progname);
        exit(0);
    }

    
#ifndef WIN32
    if (isatty(STDIN_FILENO) && !show_all_meat) {
        if (signal(SIGWINCH, sig_winch) == SIG_ERR)
            perror("cannot catch SIGWINCH. terminal will not resize.");
        if (ioctl(STDIN_FILENO, TIOCGWINSZ, (void *)&size) < 0)
            perror("cannot get terminal size. reverting to default.");
        else
            gWidth = size.ws_col;
    }
#endif

    if ((err = TM_Select_Domain(&sid, argv[0], argv[1], argv[2], "packet")) != 0)
        do_error(sid, err);

    if (argv[3])
        strcpy(stream, argv[3]);

    if ((err = TM_Select_Stream(sid, stream)) != 0)
        do_error(sid, err);

    /* if ((err = TM_Select_Stream_TimeRange(sid, 12079.5, 12079.6)) != 0) */
    /*     do_error(sid, err); */

    for (;;) {
        if ((err = TM_Find_Event(sid)) != 0)
            do_error(sid, err);

        if ((err = TM_Get_Item_R8(sid, "CCSDS_SCET_UR8", doublebuf, MAXBUFSIZE, &numvals)) != 0)
            do_error(sid, err);
        if ((err = TM_UR8_to_String(doublebuf[0], mesg)) != 0)
            do_error(sid, err);

        printf("%06d %s ", count, mesg);

        if ((err = TM_Get_Item_I4(sid, "CCSDS_HEADER_BYTES", intbuf, MAXBUFSIZE, &numvals)) != 0)
            do_error(sid, err);

        /* XXX this will break if the terminal doesn't support color, most do */
        for (j = 0; j < numvals; j++)
            printf("%c[%dm%02x%c[0m ", 0x1B, j==1?1:0, intbuf[j], 0x1B);

        if ((err = TM_Get_Item_I4(sid, "CCSDS_MEAT", intbuf, MAXBUFSIZE, &numvals)) != 0)
            do_error(sid, err);

        printf("%c[1m", 0x1B);
#ifndef WIN32
        if (show_all_meat) {
            for (j = 0; j < numvals; j++)
                printf("%02x ", intbuf[j]);
        } else if (gWidth <= 82) {
#endif
            /* 82 since we don't add a column of meat until we hit 83 columns */
            printf("%02x..%02x %02x ", intbuf[0], intbuf[numvals - 2], intbuf[numvals - 1]);
#ifndef WIN32
        } else {
            /* for every 3 columns of additional width, we add one more byte of meat */
            bytes_of_meat = floor((float)(gWidth - 80) / 3.0);

            if (bytes_of_meat < numvals)
            {
                for (j = 0; j < bytes_of_meat && j < numvals - 2; j++)
                    printf("%02x ", intbuf[j]);
                /* backspace over the last space so the line comes out looking nice */
                printf("\b..%02x %02x ", intbuf[numvals - 2], intbuf[numvals - 1]);
            }
            else
            {
                for (j = 0; j < numvals; j++)
                    printf("%02x ", intbuf[j]);
                printf("\b");
            }                
        }
#endif
        printf("%c[0m", 0x1B);

        if (show_cksum)
        {
            if ((err = TM_Get_Item_I4(sid, "CCSDS_ENTIRE_PACKET", intbuf, MAXBUFSIZE, &numvals)) != 0)
                do_error(sid, err);

            chksum = 0;
            for (j = 0; j < numvals; j++)
                chksum += intbuf[j];

            if ((chksum & 0xFF) == 0)
                printf("(%02x)\n", chksum & 0xFF);
            else
                printf("(%c[0;31m%02x%c[0m)\n", 0x1B, chksum & 0xFF, 0x1B);
        }
        else
        {
            printf("\n");
        }

        fflush(stdout);
        /* roll the counter over */
        count == 999999 ? count = 0 : count++;
    }
    TM_Close(sid);
    return (0);
}

#ifndef WIN32
void
sig_winch(int signal)
{
    struct winsize size;
    
    if (ioctl(STDIN_FILENO, TIOCGWINSZ, (void *)&size) < 0)
        perror("cannot get terminal size");
    else
        gWidth = size.ws_col;
}
#endif

void
do_error(uint32_t sid, int32_t err)
{
    char           *mesg = NULL;
    int32_t         numvals;
    int             isc;

    if (err == 0)
        return;

    mesg = malloc(MAXBUFSIZE);
    if (!mesg)
        return;

    if ((isc = TM_Error_Stack(sid, "name", mesg, MAXBUFSIZE, &numvals)) == 0)
        (void) fprintf(stderr, "Error Name: %s\n", mesg);
    else
        (void) fprintf(stderr, "Fatal error. Cound not get Error Stack.\n");
    if ((isc = TM_Error_Stack(sid, "message", mesg, MAXBUFSIZE, &numvals)) == 0)
        (void) fprintf(stderr, "Error Message: %s\n", mesg);
    else
        (void) fprintf(stderr, "Fatal error. Cound not get Error Stack.\n");
    if ((isc = TM_Error_Stack(sid, "description", mesg, MAXBUFSIZE, &numvals)) == 0)
        (void) fprintf(stderr, "Error Description: %s\n", mesg);
    else
        (void) fprintf(stderr, "Fatal error. Cound not get Error Stack.\n");
    if ((isc = TM_Error_Stack(sid, "code", mesg, MAXBUFSIZE, &numvals)) == 0)
        (void) fprintf(stderr, "Error Code: %s\n", mesg);
    else
        (void) fprintf(stderr, "Fatal error. Cound not get Error Stack.\n");
    (void) fflush(stderr);

    if (err < 0) {
        TM_Close(sid);
        exit(-1);
    }

    if (mesg)
        free(mesg);
}

void
usage(const char *progname)
{
    fprintf(stderr, "usage: %s [-c] [-h] [-a] Mission Spacecraft Instrument\n", progname);
    fprintf(stderr, "-h\t\tThis message.\n");
    fprintf(stderr, "-c\t\tUse chooser to select stream.\n");
    fprintf(stderr, "-a\t\tShow all CCSDS meat.\n");
    fprintf(stderr, "-k\t\tShow checksum.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "example:\n");
    fprintf(stderr, "$ packets SPP SPP FIELDS\n");
    fprintf(stderr, "$ packets -k STEREO STEREO_A SWAVES\n");
}
